{~~ This unit contains the TIB_CursorGrid component. }
unit IB_CursorGrid;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids,

  IB_Header,
  IB_Components;

type
{~~...
This component acts as a buffer of fetched data.}
  TIB_CursorGrid = class(TStringGrid)
  private
  { Property Storage }
    FIB_DataLink: TIB_DataLink;
    FFrozen: boolean;
    FMovingRow: boolean;
  { Property Access Methods }
    procedure SetIB_DataSource( AValue: TIB_DataSource); virtual;
    function GetIB_DataSource: TIB_DataSource; virtual;
    procedure SetFrozen( AValue: boolean ); virtual;
  { IB_CursorLink methods }
    procedure IB_DataSetStateChanged( Sender: TIB_DataLink;
                                      IB_DataSource: TIB_DataSource );
    procedure IB_DataSetActiveChanged( Sender: TIB_DataLink;
                                       IB_DataSource: TIB_DataSource );
    procedure IB_CursorRowDataChanged( Sender: TIB_DataLink;
                                       IB_DataSource: TIB_DataSource;
                                       IB_Column: TIB_Column );
    procedure IB_CursorRowUpdateData( Sender: TIB_DataLink;
                                      IB_DataSource: TIB_DataSource;
                                      IB_Column: TIB_Column );
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
  protected
  { Inherited Methods }
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
  { Utility Methods }
    procedure UpdateSQLCursorGrid;
    procedure CloseSQLCursorGrid;
    procedure ClearSQLCursorGrid;
    procedure FillSQLCursorGrid;
    procedure PostSQLCursorGrid;
    procedure RefreshSQLCursorGrid;
    procedure UpdateColumnWidths;
  { System Methods }
    procedure SysDataChanged;
    procedure SysUpdateData;
  public
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;

{~~...
Refresh grid when the data of the underlying dataset changes.

This happens automatically if Frozen is false.}
    procedure UpdateGrid;


{~~$<*#>...
This property freezes the grid so that is does not respond to changes in ...
the underlying dataset.

This can be useful if the user fetches all the rows and then wants ...
to be able to double-click on any of the fetched rows and have that row ...
re-fetched for editing without clearing the previously fetched batch of rows.

This could be accomplished by setting Frozen to true just after the last ...
row is fetched. Frozen could be set to false just prior to another ...
request to go to the last record in a selected dataset. This way the grid ...
would always represent a complete sub-set of data. Then, using this sub-set ...
the user can select and work with one record at a time by double-clicking ...
on them in the grid.}
    property Frozen: boolean read FFrozen write SetFrozen;

  published

{~~$<*>...
Reference to the DataSource for the grid.}
    property DataSource: TIB_DataSource read GetIB_DataSource
                                        write SetIB_DataSource;
  end;

implementation

constructor TIB_CursorGrid.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FIB_DataLink := TIB_DataLink.Create( Self );
  with FIB_DataLink do begin
    OnStateChanged := IB_DataSetStateChanged;
    OnActiveChanged := IB_DataSetActiveChanged;
    OnDataChanged := IB_CursorRowDataChanged;
    OnUpdateData := IB_CursorRowUpdateData;
  end;
  Options := Options - [goEditing];
end;

destructor TIB_CursorGrid.Destroy;
begin
  DataSource := nil;
  with FIB_DataLink do begin
    OnStateChanged := nil;
    OnDataChanged  := nil;
    OnUpdateData := nil;
  end;
  inherited Destroy;
end;

procedure TIB_CursorGrid.Loaded;
begin
  inherited Loaded;
  UpdateGrid;
end;

{------------------------------------------------------------------------------}

function TIB_CursorGrid.GetIB_DataSource: TIB_DataSource;
begin
  if FIB_DataLink <> nil then begin
    Result := FIB_DataLink.DataSource;
  end else begin
    Result := nil;
  end;
end;

procedure TIB_CursorGrid.SetIB_DataSource( AValue: TIB_DataSource);
begin
  if FIB_DataLink <> nil then begin
    if FIB_DataLink.DataSource <> AValue then begin
      FIB_DataLink.DataSource := AValue;
      UpdateGrid;
    end;
  end;
end;

procedure TIB_CursorGrid.SetFrozen( AValue: boolean );
begin
  if Frozen <> AValue then begin
    FFrozen := AValue;
    if Frozen = false then begin
      UpdateGrid;
    end;
  end;
end;

{------------------------------------------------------------------------------}

procedure TIB_CursorGrid.IB_DataSetStateChanged( Sender: TIB_DataLink;
                                               IB_DataSource: TIB_DataSource);
begin
  UpdateGrid;
end;

procedure TIB_CursorGrid.IB_DataSetActiveChanged( Sender: TIB_DataLink;
                                              IB_DataSource: TIB_DataSource );
begin
  RefreshSQLCursorGrid;
end;

procedure TIB_CursorGrid.IB_CursorRowDataChanged( Sender: TIB_DataLink;
                                                IB_DataSource: TIB_DataSource;
                                                IB_Column: TIB_Column);
begin
  try
    FMovingRow := true;
    SysDataChanged;
    if IB_DataSource <> nil then begin
      if IB_DataSource.IB_DataSet <> nil then begin
        with IB_DataSource.IB_DataSet do begin
          if not Unidirectional and
             ( RowNum > 0 ) and
             ( RowNum < Self.RowCount ) and
             ( RowNum <> Row ) then begin
            Row := RowNum;
          end;
        end;
      end;
    end;
  finally
    FMovingRow := false;
  end;
end;

procedure TIB_CursorGrid.IB_CursorRowUpdateData( Sender: TIB_DataLink;
                                                IB_DataSource: TIB_DataSource;
                                                IB_Column: TIB_Column );
begin
  SysUpdateData;
end;

{------------------------------------------------------------------------------}

procedure TIB_CursorGrid.UpdateGrid;
begin
  if Frozen or ( DataSource = nil ) then Exit;
  if not ( csLoading    in ComponentState ) and
     not ( csDestroying in ComponentState ) then begin
    case DataSource.DataSetState of
      dsInactive: CloseSQLCursorGrid;
      dsPrepared: ClearSQLCursorGrid;
    end;
    UpdateSQLCursorGrid;
  end;
end;

procedure TIB_CursorGrid.SysDataChanged;
begin
  if Frozen or ( DataSource = nil ) then Exit;
  with DataSource do begin
    if IB_DataSet <> nil then with IB_DataSet do begin
      if IB_DataSet is TIB_BDataSet then begin
        with IB_DataSet as TIB_BDataSet do begin
          if Self.RowCount > BufferRowCount + FixedRows + 1 then begin
            Self.RowCount := BufferRowCount + FixedRows + 1;
          end;
        end;
      end;
      if BOF then begin
        ClearSQLCursorGrid;
      end else if EOF and Unidirectional then begin
      end else begin
        FillSQLCursorGrid;
      end;
    end;
  end;
end;

procedure TIB_CursorGrid.SysUpdateData;
begin
  if Frozen or ( DataSource = nil ) then Exit;
// PostSQLCursorGrid;  // Not ready yet!
// Need to isolate a modified flag for each cell.
end;

{------------------------------------------------------------------------------}

function TIB_CursorGrid.SelectCell( ACol, ARow: longint ): boolean;
begin
  if FMovingRow then begin
    Result := true;
  end else begin
    Result := inherited SelectCell( ACol, ARow );
    if Result and
      ( DataSource <> nil ) and
      ( DataSource.IB_DataSet <> nil ) then begin
      with DataSource.IB_DataSet do begin
        if Active and not Unidirectional and
          ( Row > 0 ) and
          ( ARow <> RowNum ) then begin
          RowNum := ARow;
          Result := false;
        end;
      end;
    end;
  end;
end;

{------------------------------------------------------------------------------}
{   Grid handling procedures                                                   }
{------------------------------------------------------------------------------}

procedure TIB_CursorGrid.UpdateSQLCursorGrid;
var
  ii: integer;
begin
  if ( FIB_DataLink.Prepared ) then begin
    with DataSource.IB_DataSet do begin
      ColCount := OutputCount;
      for ii := 0 to OutputCount - 1 do begin
        Cells[ ii, 0 ] := OutputRow[ ii ].DisplayName;
      end;
    end;
  end else begin
    CloseSQLCursorGrid;
  end;
end;

procedure TIB_CursorGrid.CloseSQLCursorGrid;
var
  ii: integer;
begin
  RowCount := 2;
  FixedRows := RowCount - 1;
  RowCount := FixedRows + 1;
  for ii := 0 to FixedRows do begin
    Rows[ ii ].Clear;
  end;
  ColCount := 1;
end;

procedure TIB_CursorGrid.ClearSQLCursorGrid;
var
  ii: integer;
  OldCursor: TCursor;
begin
  OldCursor := Screen.Cursor;
  try
    Screen.Cursor := crHourglass;
    with DataSource.IB_DataSet do begin
      for ii := FixedRows to RowCount - 1 do begin
        Rows[ ii ].Clear;
      end;
      Self.RowCount := FixedRows + 1;
    end;
    UpdateColumnWidths;
  finally
    Screen.Cursor := OldCursor;
  end;
end;

procedure TIB_CursorGrid.FillSQLCursorGrid;
var
  Index: integer;
begin
  if ( FIB_DataLink.Active ) then begin
    with DataSource.IB_DataSet do begin
      if Self.RowCount < RowNum + FixedRows then begin
        Self.RowCount := RowNum + FixedRows;
      end;
      for Index := 0 to OutputCount - 1 do begin
        if BOF then begin
        end else if EOF then begin
          Cells[ Index, RowNum + FixedRows - 1] := '';
        end else if OutputRow[ Index ].IsNull then begin
          Cells[ Index, RowNum + FixedRows - 1] := '<NULL>';
        end else if OutputRow[ Index ].IsBlob then begin
          Cells[ Index, RowNum + FixedRows - 1] := '(BLOB)';
        end else begin
          Cells[ Index, RowNum + FixedRows - 1] := OutputRow[ Index ].AsString;
        end;
      end;
    end;
    RefreshSQLCursorGrid;
  end;
end;

procedure TIB_CursorGrid.RefreshSQLCursorGrid;
var
  Index: integer;
begin
  if ( FIB_DataLink.Active ) then begin
    if DataSource.IB_DataSet is TIB_BDataSet then begin
      with DataSource.IB_DataSet as TIB_BDataSet do begin
        BufferRowNum := TopRow;
        while not BufferEOF do begin
          if Self.RowCount < BufferRowNum + FixedRows then begin
            Self.RowCount := BufferRowNum + FixedRows;
          end;
          for Index := 0 to OutputCount - 1 do begin
            if BufferRow[ Index ].IsNull then begin
              Cells[ Index, BufferRowNum + FixedRows - 1 ] := '<NULL>';
            end else if BufferRow[ Index ].IsBlob then begin
              Cells[ Index, BufferRowNum + FixedRows - 1 ] := '(BLOB)';
            end else begin
              Cells[ Index, BufferRowNum + FixedRows - 1 ] :=
                BufferRow[ Index ].AsString;
            end;
          end;
          if (( BufferRowNum - TopRow ) > VisibleRowCount ) then begin
            Break;
          end else begin
            BufferRowNum := BufferRowNum + 1;
          end;
        end;
      end;
    end;
  end;
end;

procedure TIB_CursorGrid.PostSQLCursorGrid;
var
  Index: integer;
begin
  if ( FIB_DataLink.Active ) then begin
    with DataSource.IB_DataSet do begin
      if Self.RowCount < RowNum + FixedRows then begin
        Self.RowCount := RowNum + FixedRows;
      end;
      for Index := 0 to OutputCount - 1 do begin
        if BOF then begin
        end else if EOF then begin
          Cells[ Index, RowNum + FixedRows - 1] := '';
        end else if OutputRow[ Index ].IsNull then begin
          Cells[ Index, RowNum + FixedRows - 1] := '<NULL>';
        end else if OutputRow[ Index ].IsBlob then begin
          Cells[ Index, RowNum + FixedRows - 1] := '(BLOB)';
        end else if Cells[ Index, RowNum + FixedRows - 1] <>
                    OutputRow[ Index ].AsString then begin
          OutputRow[ Index ].AsString := Cells[ Index, RowNum + FixedRows - 1];
        end;
      end;
    end;
  end;
end;

procedure TIB_CursorGrid.UpdateColumnWidths;
var
  ii: integer;
  DataWidth: integer;
  HeaderWidth: integer;
  NewWidth: integer;
begin
  if ( FIB_DataLink.Prepared ) then begin
    with DataSource.IB_DataSet do begin
      ColCount := OutputCount;
      for ii := 0 to OutputCount - 1 do with OutputRow[ ii ] do begin
        HeaderWidth := Length( Trim( DisplayName ));
        if ( SQLType = SQL_DATE  ) or
           ( SQLType = SQL_DATE_ ) then begin
          DataWidth := 16;
        end else begin
          DataWidth := SQLLen;
        end;
        if HeaderWidth > DataWidth then begin
          NewWidth := HeaderWidth * 8;
        end else begin
          NewWidth := DataWidth * 8;
        end;
        if NewWidth < 70 then begin
          NewWidth := 70;
        end;
        if NewWidth > 200 then begin
          NewWidth := 200;
        end;
        ColWidths[ ii ] := NewWidth;
        ColWidths[ ii ] := NewWidth;
      end;
    end;
  end;
end;

procedure TIB_CursorGrid.CMEnter( var Message: TCMEnter );
begin
  FIB_DataLink.SetFocus;
  inherited;
end;

procedure TIB_CursorGrid.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown( Key, Shift );
  if FIB_DataLink.Active then with DataSource.IB_DataSet do begin
    if ssCtrl in Shift then case Key of
      VK_HOME: begin
        First;
        Key := 0;
      end;
      VK_END: begin
        Last;
        Key := 0;
      end;
      VK_DOWN: begin
        Next;
        Key := 0;
      end;
      VK_UP: begin
        Prior;
        Key := 0;
      end;
    end;
  end;
end;

end.


