{~~ This unit encapsulates the InterBase event functionality.}
unit IB_Events;

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs,

  IB_Header,
  IB_Components;

const
{~~ Maximum number of events that can be associated in one block.}
MaxEvents   = 15;

{~~ Maximum length of an event name.}
EventLength = 64;

type
{~~ Event type for the notification received from the database.}
TEventAlert = procedure( Sender: TObject;
                         EventName: string;
                         EventCount: longint ) of object;

{~~ Structure used to pass the event block info to the IB API.}
TEventBuffer = array[ 0..MaxEvents   - 1,
                      0..EventLength - 1 ] of char;

{~~ This component is used to encapsulate InterBase's event functionality. }
TIB_Events = class( TComponent)
  private
    FIB_ConnectionLink: TIB_ConnectionLink;
    FEvents: TStrings;
    Buffer: TEventBuffer;
    CS: TRTLCriticalSection;
    ResultBuffer: PChar;
    EventBuffer: PChar;
    EventBufferLen: integer;
    EventID: isc_long;
    EventThreadID: integer;
    function GetIB_Connection: TIB_Connection; virtual;
    procedure SetIB_Connection( AValue: TIB_Connection ); virtual;
    procedure CheckConnection; virtual;
    function GetIB_Session: TIB_Session; virtual;
    function GetPdbHandle: pisc_db_handle; virtual;
    procedure SetEvents( AValue: TStrings ); virtual;
    procedure SetRegistered( AValue: boolean ); virtual;
    function GetRegistered: boolean; virtual;
    function GetQueued: boolean; virtual;
  private
  { Utility Methods }
    SkipNextAlert: boolean;
    ProcessingEvents: boolean;
    RegisteredState: boolean;
    FOnEventAlert: TEventAlert;
    FBeforeRegister: TNotifyEvent;
    FBeforeUnregister: TNotifyEvent;
    FAfterRegister: TNotifyEvent;
    FAfterUnregister: TNotifyEvent;
    procedure API_QueueEvents;
    procedure EventChange( Sender: TObject );
    procedure UpdateResultBuffer( length: short; updated: PChar );
    procedure BeforeConnectionAssignment( OldLink: TIB_ConnectionLink;
                                          Old: TIB_Connection );
  protected
  { Utility Methods }
    procedure HandleEvent;
    procedure Loaded; override;
  public
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
  { Methods }
{~~...
Remove events from the database queue to ignore event notifications.
}   procedure CancelEvents;
{~~...
 Enter events in the database queue to receive event notifications.}
    procedure QueueEvents;
{~~$<*>...
Register events in the database in order to queue them.
This method automatically calls the QueueEvents method.
}
    procedure RegisterEvents;
{~~$<*>...
Remove events from the database event queue and registry.
}   procedure UnregisterEvents;
  { Properties }
{~~...
Reference to the IB_Session used for the current thread.
}   property IB_Session: TIB_Session read GetIB_Session;
{~~...
Native InterBase connection handle.
}   property PdbHandle: pisc_db_Handle read GetPdbHandle;
{~~...
Property to tell if the events are entered into the database event queue.
}   property Queued: boolean read GetQueued;
  published
{~~$<*>...
Reference to the database which will be generating event notifications.
}   property IB_Connection: TIB_Connection read GetIB_Connection
                                           write SetIB_Connection;
{~~$<*>...
List of event names to be monitored and receive notifiaction for.

If this property is changed the currently defined events are automatically...
cancelled and unregistered.
}   property Events: TStrings read FEvents
                              write SetEvents;
{~~$<*>...
Gets or sets the registered status of the events.
}   property Registered: boolean read GetRegistered
                                 write SetRegistered;
{~~$<*>...
Event where notification is received from the databse.
It is handled in a seperate thread asynchronously so use caution.
}   property OnEventAlert: TEventAlert read FOnEventAlert
                                       write FOnEventAlert;
{~~$<!>}
    property BeforeRegister: TNotifyEvent read FBeforeRegister
                                          write FBeforeRegister;
{~~$<!>}
    property BeforeUnregister: TNotifyEvent read FBeforeUnregister
                                            write FBeforeUnregister;
{~~$<!>}
    property AfterRegister: TNotifyEvent read FAfterRegister
                                         write FAfterRegister;
{~~$<!>}
    property AfterUnregister: TNotifyEvent read FAfterUnregister
                                           write FAfterUnregister;
  end;

implementation

{------------------------------------------------------------------------------}

procedure HandleEvent( param: integer ); stdcall;
begin
  // don't let exceptions propogate out of thread
  try
    TIB_Events( param ).HandleEvent;
  except
    Application.HandleException( nil );
  end;
end;

procedure IBEventCallback( ptr: pointer; length: short; updated: PChar ); cdecl;
var
  ThreadID: integer;
begin
  // Handle events asynchronously in second thread
  EnterCriticalSection( TIB_Events( ptr ).CS );
  TIB_Events( ptr ).UpdateResultBuffer( length, updated );
  CloseHandle( CreateThread( nil, 8192, @HandleEvent, ptr, 0, ThreadID ));
  LeaveCriticalSection( TIB_Events( ptr ).CS );
end;

{------------------------------------------------------------------------------}

constructor TIB_Events.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  InitializeCriticalSection( CS );
  FIB_ConnectionLink := TIB_ConnectionLink.Create( Self );
  with FIB_ConnectionLink do begin
    BeforeAssignment := BeforeConnectionAssignment;
    BeforeDisconnect := BeforeConnectionAssignment;
  end;
  FEvents := TStringList.Create;
  with FEvents as TStringList do begin
    OnChange := EventChange;
    Duplicates := dupIgnore;
  end;
end;

destructor TIB_Events.Destroy;
begin
  with FIB_ConnectionLink do begin
    BeforeAssignment := nil;
    BeforeDisconnect := nil;
  end;
  IB_Connection := nil;
  with FEvents as TStringList do begin
    OnChange := nil;
  end;
  Destroying;
  FEvents.Free;
  FEvents := nil;
  DeleteCriticalSection( CS );
  inherited Destroy;
end;

procedure TIB_Events.Loaded;
begin
  inherited Loaded;
  try
    if RegisteredState then RegisterEvents;
  except
    if csDesigning in ComponentState then
      Application.HandleException( Self )
    else raise;
  end;
end;

{------------------------------------------------------------------------------}

function TIB_Events.GetIB_Session: TIB_Session;
begin
  if IB_Connection <> nil then begin
    Result := IB_Connection.IB_Session;
  end else begin
    Result := nil;
  end;
end;

function TIB_Events.GetPdbHandle: pisc_db_handle;
begin
  if IB_Connection <> nil then begin
    Result := IB_Connection.PdbHandle;
  end else begin
    Result := nil;
  end;
end;

function TIB_Events.GetIB_Connection: TIB_Connection;
begin
  Result := FIB_ConnectionLink.IB_Connection;
end;

procedure TIB_Events.SetIB_Connection( AValue: TIB_Connection );
begin
  FIB_ConnectionLink.IB_Connection := AValue;
end;

function TIB_Events.GetRegistered: boolean;
begin
  Result := EventBufferLen <> 0;
end;

function TIB_Events.GetQueued: boolean;
begin
  Result := EventID <> 0;
end;

procedure TIB_Events.SetEvents( AValue: TStrings );
begin
  FEvents.Assign( AValue );
end;

procedure TIB_Events.SetRegistered( AValue: boolean );
begin
  if ( csReading in ComponentState ) then begin
    RegisteredState := AValue
  end else if ( Registered <> AValue ) and AValue then begin
    RegisterEvents
  end else begin
    UnregisterEvents;
  end;
end;

{------------------------------------------------------------------------------}

procedure TIB_Events.CheckConnection;
begin
  if ( IB_Connection = nil ) then begin
    raise EIB_Error.Create( 'IB_Connection is not assigned' );
  end;
  IB_Connection.Connected := true;
  if not IB_Connection.Connected then begin
    raise EIB_Error.Create( 'IB_Connection is not connected' );
  end;
end;

procedure TIB_Events.BeforeConnectionAssignment( OldLink: TIB_ConnectionLink;
                                                 Old: TIB_Connection );
begin
  UnregisterEvents;
end;

procedure TIB_Events.EventChange( Sender: TObject );
begin
  // check for blank event
  if TStringList( Events ).IndexOf( '' ) <> -1 then
    raise EIB_Error.Create( 'Invalid event' );
  // check for too many events
  if Events.Count > MaxEvents then begin
    TStringList( Events ).OnChange := nil;
    Events.Delete( MaxEvents );
    TStringList( Events ).OnChange := EventChange;
    raise EIB_Error.Create( 'Maximum events' );
  end;
  if Registered then begin
    RegisterEvents;
  end;
end;

{------------------------------------------------------------------------------}

procedure TIB_Events.HandleEvent;
var
  ii: integer;
  status: status_vector;
begin
  try
    // prevent modification of vital data structures while handling events
    EnterCriticalSection( CS );
    ProcessingEvents := true;
    if EventBuffer <> nil then begin
      IB_Session.isc_event_counts( @status,
                                   EventBufferLen,
                                   EventBuffer,
                                   ResultBuffer );
      if Assigned( FOnEventAlert ) and not SkipNextAlert then begin
        for ii := 0 to Events.Count - 1 do begin
          try
            if ( status[ ii ] <> 0 ) then begin
              FOnEventAlert( Self,
                             Events[ Events.Count - ii - 1 ],
                             status[ ii ] );
            end;
          except
            Application.HandleException( Self );
          end;
        end;
      end;
      API_QueueEvents;
      SkipNextAlert := false;
    end;
  finally
    ProcessingEvents := false;
    LeaveCriticalSection( CS );
  end;
end;

{------------------------------------------------------------------------------}

procedure TIB_Events.API_QueueEvents;
begin
  with IB_Session do begin
    errCode := isc_que_events( @status,
                               PdbHandle,
                               @EventID,
                               EventBufferLen,
                               EventBuffer,
                               isc_callback( @IBEventCallback ),
                               Self );
    if errcode <> 0 then HandleException( Self, nil );
  end;
  SkipNextAlert := true;
end;

procedure TIB_Events.CancelEvents;
begin
  if ProcessingEvents then raise EIB_Error.Create( 'Processing events'#13#10 +
                                                   'Cannot cancel events' );
  if Queued then begin
    try
      // wait for event handler to finish before cancelling events
      EnterCriticalSection( CS );
      with IB_Session do begin
        errcode := isc_cancel_events( @status, PdbHandle, @EventID );
        if ( ErrCode = isc_lost_db_connection ) or
           ( ErrCode = isc_network_error ) then begin
          // Throw events away! Connection was lost!
          ErrCode := 0;
        end;
        if errcode <> 0 then HandleException( Self, nil );
        EventID := 0;
      end;
    finally
      LeaveCriticalSection( CS );
    end;
  end;
end;

procedure TIB_Events.QueueEvents;
begin
  if not Registered   then raise EIB_Error.Create( 'No events registered' );
  if ProcessingEvents then raise EIB_Error.Create( 'Processing events'#13#10 +
                                                   'Cannot queue events' );
  if not Queued then begin
    try
      // wait until current event handler is finished before queuing events
      EnterCriticalSection( CS );
      API_QueueEvents;
    finally
      LeaveCriticalSection( CS );
    end;
  end;
end;

procedure TIB_Events.RegisterEvents;
var
  ii: integer;
  bufptr: pointer;
  eventbufptr: pointer;
  resultbufptr: pointer;
  buflen: integer;
  isc_event_block: Tisc_event_block;
begin
  if Registered then begin
    UnregisterEvents;
  end else begin
    if not ( csDestroying in ComponentState ) then CheckConnection;
  end;
  if Assigned( FBeforeRegister ) then begin
    FBeforeRegister( Self );
  end;
  if Events.Count > 0 then begin
    for ii := 0 to Events.Count - 1 do begin
      StrPCopy( @Buffer[ii][0], Events[ii] );
    end;
    ii              := Events.Count;
    bufptr          := @buffer[0];
    eventbufptr     := @EventBuffer;
    resultBufPtr    := @ResultBuffer;
    isc_event_block := IB_Session.isc_event_block;
    asm
      mov ecx, dword ptr [ii]
      mov eax, dword ptr [bufptr]
      @@1:
      push eax
      add  eax, EventLength
      loop @@1
      push dword ptr [ii]
      push dword ptr [resultBufPtr]
      push dword ptr [eventBufPtr]
      call [isc_event_block]
      mov  dword ptr [bufLen], eax
      mov eax, dword ptr [ii]
      shl eax, 2
      add eax, 12
      add esp, eax
    end;
    EventBufferlen := Buflen;
    QueueEvents;
    if Assigned( FAfterRegister ) then begin
      FAfterRegister( Self );
    end;
  end;
end;

procedure TIB_Events.UnregisterEvents;
begin
  if ProcessingEvents then raise EIB_Error.Create( 'Processing events'#13#10 +
                                                   'Cannot unregister' );
  if Assigned( FBeforeUnregister ) then begin
    FBeforeUnregister( Self );
  end;
  if Registered then begin
    CancelEvents;
    IB_Session.isc_free( EventBuffer  );
    EventBuffer := nil;
    IB_Session.isc_free( ResultBuffer );
    ResultBuffer := nil;
    EventBufferlen := 0;
    if Assigned( FAfterUnregister ) then begin
      FAfterUnregister( Self );
    end;
  end;
end;

procedure TIB_Events.UpdateResultBuffer( Length: short; Updated: PChar );
var
  ii: integer;
begin
  for ii := 0 to Length - 1 do begin
    ResultBuffer[ ii ] := Updated[ ii ];
  end;
end;

end.
