{
 class TMemoryMappedFileStream
 Encapsulation of the Windows Memory Mapped object.

 Use, modify or distribute (do not sell) as you please
 as long as you give me a credit.

 Allows for
        1) simple interprocess communications via the named
           mapped objects transparently accessible either
           as the TStream decendents or as arrays using
           Memory property
        2) Access to any files up to 4GB size as the memory arrays
           using Memory property

 Usage:
        Just like any TStream descendent ( Read(), Write(),
        Size, Position, etc.)

 Additional properties and methods:
   procedure Flush   ; flushes MMF contents to the disk
   property Memory   :pointer; (read only) - use like any
                               other pointer, DO NOT use in the
                               memory de/re/allocation routines.
   property FileName :string; (read only) - name of the
                              file mapped to memory
   property MapName  :string; (read only) - name of the memory
                              mapped object (has nothing to do with
                              the FileName).
   property MemoryDelta:longint; (read/write) - increment with which
                                 object's allocated memory will grow

 Hints:

  To create new mapped file object, use constructor
    Create(FileName, MapName,Mode)
  FileName - name of the file to map, can be '' if you want to use
             Windows swap file (useful for interprocess communications)
  MapName - name of the mapped object to create, can be '' if
            you do not want to open it using different mapped file object
  Mode - file open/creat mode, see TFileStream.Create help

  To use existing mapped file object, use constructor
     Open(MapName, Mode)
  MapName - name of the existing mapped file object (see above),
            creator can reside in a different process.
  Mode - Mapped object open mode, see TFileStream.Create help.
         DO NOT use fmCreate mode, exception will be thrown!
         All other modes are Ok.

Example - count all letters 'a' in a file
         type hugearray=array[0..$FFFFFFFF-1] of char;
         var mmf: TMemoryMappedFileStream;
             HugeFileName:string;
             i,counter:longint;
         begin
         ....
         HugeFileName:=...;
         try
          mmf:=TMemoryMappedFileStream.Create(HugeFileName,'',fmOpenReadWrite);
          counter:=0;
          for i:=0 to mmf.Size-1 do
           if hugearray(mmf.memory^)[i]='a' then inc(counter);
         finally
          mmf.Destroy;
         end;
         ....


Warning
       As soon as a named memory mapped object is created and open
       by using a second handle you cannot change its size, e.g.,
       ....
       var mmf1,mmf2: TMemoryMappedFileStream;
       ....
       //Create new memory mapped object, defaul size 64kB
       mmf1:=TMemoryMappedFileStream.Create('c:\temp\mmf.tmp','mem_map',fmCreate);
       ....
       //Open it either from the same process or from a different one
       mmf2:=TMemoryMappedFileStream.Open('mem_map',fmOpenReadWrite);
       ....
       //Try setting MM object's size to 1Mb
       mmf1.Size:=$100000;
       //Exception is thrown...

       Remember that Size can be changed both explicitly (above) or
       implicitly by calling Write(Something,SizeOf(Something)), which
       in case of insufficient memory tries to allocate more memory.

       A workaround is to set stream's size BEFORE opening the MM object:
       ....
       var mmf1,mmf2: TMemoryMappedFileStream;
       ....
       mmf1:=TMemoryMappedFileStream.Create('c:\temp\mmf.tmp','mem_map',fmCreate);
       //mmf2 is not open yet, you can change size
       mmf1.Size:=$100000;
       ....
       mmf2:=TMemoryMappedFileStream.Open('mem_map',fmOpenReadWrite);
       ....


       Another thing is that for some reason Windows does not like
       creating or opening MMF with a read-only (fmOpenRead) attribute.
       If anybody knows why I would really appreciate explaining it to me.


       Have fun!


      Dmitry Streblechenko
      Department of Physics and Astronomy
      Arizona State University
      Box 871504
      Tempe, AZ 85287-1504

      E-mail: dmitrys@phyast.la.asu.edu



}

unit MemoryMappedFiles;

interface

uses Classes,Windows,SysUtils,Dialogs;

type

 TMemoryMappedFileStream = class (TStream)
  private
   FFileHandle:THandle;
   FMapHandle:THandle;
   FFileName:string;
   FMemory:pointer;
   FCapacity,FSize,FPosition:longint;
   //file access attributes
   FAccess:longint;
   //file sharing attributes
   FShare:longint;
   //mapping object page protection attributes
   FPageProtect:longint;
   //protection of the mapped memory
   FMemProtect:longint;
   //Name of the file mapping object
   FMapName:string;
   //file size increment
   FMemoryDelta:longint;
   procedure SetCapacity(NewCapacity:longint);
   procedure SetAttributes(Mode:word);
   procedure CloseHandles;
  public
   constructor Create(const FileName, MapName: string; Mode: Word);
   constructor Open(const MapName:string; Mode:word);
   destructor Destroy; override;
   function Read(var Buffer;Count:longint):longint;override;
   function Write(const Buffer;Count:longint):longint;override;
   procedure SetSize(NewSize:longint);override;
   function Seek(Offset:longint;Origin:word):longint;override;
   procedure Flush;virtual;
   property Memory:pointer read FMemory;
   property FileName:string read FFileName;
   property MapName:string read FMapName;
   property MemoryDelta:longint read FMemoryDelta write FMemoryDelta;
 end;

implementation

procedure TMemoryMappedFileStream.SetAttributes(Mode:word);
begin
     if  (Mode and fmOpenReadWrite) = fmOpenReadWrite then
     begin
       FAccess :=GENERIC_WRITE OR GENERIC_READ;
       FPageProtect:=PAGE_READWRITE OR SEC_COMMIT;
       FMemProtect:=FILE_MAP_ALL_ACCESS;
     end
    else if (Mode and fmOpenWrite) = fmOpenWrite then
     begin
       FAccess :=GENERIC_WRITE OR GENERIC_READ;
       FPageProtect:=PAGE_READWRITE OR SEC_COMMIT;
       FMemProtect:=FILE_MAP_WRITE;
     end
    else  //all others including fmOpenRead=0
     begin
       FAccess :=GENERIC_READ;
       FPageProtect:=PAGE_READONLY OR SEC_COMMIT;
       FMemProtect:=FILE_MAP_READ;
     end;
   FShare:=0;
   if (Mode and fmShareDenyWrite) = fmShareDenyWrite then
         FShare:=FILE_SHARE_READ;
   if (Mode and fmShareDenyRead) = fmShareDenyRead then
         FShare:=FILE_SHARE_WRITE;
   if (Mode and fmShareExclusive) = fmShareExclusive then
         FShare:=0;
end;

constructor TMemoryMappedFileStream.Create(const FileName, MapName: string; Mode: Word);
var err:integer;
begin
  FMemoryDelta:=$FFFF;
  FFileName:=FileName;
  FMapName:=MapName;
  SetAttributes(Mode);
  if Mode = fmCreate then
  //Create new file
  begin
    FFileHandle := CreateFile(PChar(FFileName),FAccess,FShare,nil,CREATE_ALWAYS,0,0);
    if (FFileHandle = 0) then
      raise EFCreateError.CreateFmt(
          'Error creating %s'#13#10'Code %d', [FFileName,GetLastError])
    else begin
     FSize:=FMemoryDelta;
     FCapacity:=FMemoryDelta;
    end;
  end else
  //Open existing file
  begin
    FFileHandle := CreateFile(PChar(FFileName),FAccess,FShare,nil,OPEN_EXISTING,0,0);
    FSize:=GetFileSize(FFileHandle,nil);
    FCapacity:=FSize;
    if (FFileHandle = 0) or (FSize=longint($FFFFFFFF)) then
      raise EFOpenError.CreateFmt(
           'Error opening %s'#13#10'Code %d', [FFileName,GetLastError]);
  end;
  FMapHandle:=CreateFileMapping(FFileHandle,nil,FPageProtect,0,FCapacity,PChar(FMapName));
  err:=GetLastError;
  if (FMapHandle = 0) or (err <> ERROR_SUCCESS) then
      raise EFOpenError.CreateFmt('Error creating file mapping object %s'#13#10'Code %d', [FMapName,err]);
  FMemory:=MapViewOfFile(FmapHandle,FILE_MAP_WRITE,0,0,0);
  err:=GetLastError;
  if (err <> ERROR_SUCCESS) then
      raise EFOpenError.CreateFmt('Error mapping file %s'#13#10'Code %d', [FFileName,err]);
  FPosition:=0;
end;

constructor TMemoryMappedFileStream.Open(const MapName:string; Mode:word);
var err:integer;
begin
  FMemoryDelta:=$FFFF;
  FFileName:='';
  FMapName:=MapName;
  SetAttributes(Mode);
  if Mode = fmCreate then
      raise EFCreateError.Create('Use "Create" constructor with fmCreate mode');
  FSize:=FMemoryDelta;
  FCapacity:=FMemoryDelta;
  FFileHandle := 0;
  FMapHandle:=OpenFileMapping(FMemProtect,FALSE,PChar(FMapName));
  err:=GetLastError;
  if (FMapHandle = 0) or (err <> ERROR_SUCCESS) then
      raise EFOpenError.CreateFmt('Error opening file mapping object %s'#13#10'Code %d', [FMapName,err]);
  FMemory:=MapViewOfFile(FmapHandle,FILE_MAP_WRITE,0,0,0);
  err:=GetLastError;
  if err <> ERROR_SUCCESS then
      raise EFOpenError.CreateFmt('Error mapping file %s'#13#10'Code %d', [FMapName,err]);
  FPosition:=0;
end;

procedure TMemoryMappedFileStream.CloseHandles;
begin
  if FMemory <> nil then begin
   if not FlushViewOfFile(FMemory,0) then
     raise EStreamError.CreateFmt('Error flushing'#13#10'Code %d', [GetLastError]);
   if not UnmapViewOfFile(FMemory) then
     raise EStreamError.CreateFmt('Error unmapping view'#13#10'Code %d', [GetLastError]);
  end;
  if FMapHandle <> 0 then if not CloseHandle(FMapHandle) then
    raise EStreamError.CreateFmt('Error closing map %s'#13#10'Code %d', [FMapName,GetLastError]);
  if FFileHandle <> 0 then if not CloseHandle(FFileHandle) then
    raise EStreamError.CreateFmt('Error closing file %s'#13#10'Code %d', [FFileName,GetLastError]);
end;

destructor TMemoryMappedFileStream.Destroy;
begin
  CloseHandles;
end;

function TMemoryMappedFileStream.Read(var Buffer;Count:longint):longint;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Result := FSize - FPosition;
    if Result > 0 then
    begin
      if Result > Count then Result := Count;
      Move(Pointer(Longint(FMemory) + FPosition)^, Buffer, Result);
      Inc(FPosition, Result);
      Exit;
    end;
  end;
  Result := 0;
end;

function TMemoryMappedFileStream.Write(const Buffer;Count:longint):longint;
var
  Pos: Longint;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Pos := FPosition + Count;
    if Pos > 0 then
    begin
      if Pos > FSize then
      begin
        if Pos > FCapacity then
          SetCapacity(Pos);
        FSize := Pos;
      end;
      System.Move(Buffer, Pointer(Longint(FMemory) + FPosition)^, Count);
      FPosition := Pos;
      Result := Count;
      Exit;
    end;
  end;
  Result := 0;
end;


procedure TMemoryMappedFileStream.SetSize(NewSize:longint);
begin
 if NewSize < FSize then SetCapacity(NewSize)
  else if NewSize >= FSize then
     if NewSize <= FCapacity then
                               FSize:=NewSize
                              else begin
                                SetCapacity(NewSize);
                                if NewSize <= FCapacity then FSize:=NewSize;
                              end;
end;

procedure TMemoryMappedFileStream.SetCapacity(NewCapacity:longint);
var err:integer;
begin
 if NewCapacity=0 then begin
  Flush;
  FSize:=0;
  FCapacity:=0;
 end
 else begin
  CloseHandles;
  //Close MMF and try to recreate it with the new size
  //Is there a better way to change the size of an MMF?

  //FFileHandle=0 for the open (not created!)named mapping object
  if FFileHandle <> 0 then begin
   FFileHandle := CreateFile(PChar(FFileName),FAccess,FShare,nil,OPEN_EXISTING,0,0);
   err:=GetLastError;
   if (FFileHandle = 0) or (FSize=longint($FFFFFFFF)) or
      (err <> ERROR_SUCCESS) then
       raise EFOpenError.CreateFmt('Error opening %s'#13#10'Code %d', [FileName,err]);
  end
  else raise EFOpenError.CreateFmt('Cannot change size of an open mapping object %s', [FMapName]);
  FCapacity:=NewCapacity+FMemoryDelta;
  FMapHandle:=CreateFileMapping(FFileHandle,nil,FPageProtect,0,FCapacity,PChar(FMapName));
  err:=GetLastError;
  if (FMapHandle = 0) or (err <> ERROR_SUCCESS) then
      raise EFOpenError.CreateFmt('Error creating file mapping object %s'#13#10'Code %d', [FMapName,err]);
  FMemory:=MapViewOfFile(FMapHandle,FILE_MAP_WRITE,0,0,0);
  err:=GetLastError;
  if err <> ERROR_SUCCESS then
      raise EFOpenError.CreateFmt('Error mapping file %s'#13#10'Code %d', [FMapName,err]);
  FPosition:=0;
 end;
end;

function TMemoryMappedFileStream.Seek(Offset:longint;Origin:word):longint;
begin
 case Origin of
  soFromBeginning : FPosition:=Offset;
  soFromCurrent   : Inc(FPosition,Offset);
  soFromEnd       : FPosition:=FSize+Offset;
 end;
 Result:=FPosition;
end;

procedure TMemoryMappedFileStream.Flush;
begin
 FlushViewOfFile(FMemory,0);
end;

end.
