{+--------------------------------------------------------------------------+
 | Component:   TTokenList
 | Created:     12.06.97
 | Author:      Martin Waldenburg
 | Copyright    1997, all rights reserved.
 | Description: TTokenList is a general tokenizer, very fast.
 |              It contains a faster replacement for StrScan.
 | Version:     1.2
 | State:       FreeWare
 +--------------------------------------------------------------------------+}
 { Disclaimer:
   This is provided as is, expressly without a warranty of any kind.
   You use it at your own risc.}

unit mwTokenList;

{$R-}


interface

uses SysUtils, Windows, Dialogs;

type

 { TTokenList class }
  PTokenArray = ^TTokenArray;
  TTokenArray = array[0..0] of PChar;
  TTokenList = class(TObject)
  private
    Run, Start: PChar;
    FTokenPos: PChar;
    FTokenList: PTokenArray;
    FCount: Integer;
    fCapacity:Integer;
    fOrigin:PChar;
    fDelimiters: array[0..50] of Char;
    fFinished:Boolean;
    procedure Expand;
    procedure SetCapacity(NewCapacity:Integer);
    function GetToken(Index: Integer):String;
    function GetPosition(Index: Integer):Integer;
    function GetLength(Index: Integer):Integer;
    procedure SetDelimiters(value:String);
    function GetDelimiters:String;
  protected
    function GetItems(Index: Integer):PChar;
    procedure SetItems(Index: Integer;  Item: PChar);
  public
    constructor create(OriginPtr: PChar; Delim: String);
    destructor Destroy; override;
    function Add(Item: PChar):Integer;
    procedure Clear;
    procedure Pack;
    function Next:String;
    procedure Tokenize(StartCapacity: Integer);
    function ScanStr(ToScan: PChar; Sign: Char):PChar;
    property Capacity:Integer read fCapacity write SetCapacity;
    property Count: Integer read FCount;
    property Items[Index: Integer]: PChar read GetItems write SetItems; default;
    property Origin:PChar read fOrigin write fOrigin;
    property Token[Index: Integer]:String read GetToken;
    property Position[Index: Integer]:Integer read GetPosition;
    property Length[Index: Integer]:Integer read GetLength;
    property Delimiters:String read GetDelimiters write SetDelimiters;
    property Finished:Boolean read fFinished;
  published
  end;

implementation

{ TTokenList }

constructor TTokenList.create(OriginPtr: PChar; Delim: String);
begin
  inherited Create;
  fOrigin:= OriginPtr;
  Start:= FOrigin;
  Run:= Start;
  SetDelimiters(Delim);
  FFinished:= False;
end;

procedure TTokenList.Clear;
begin
  FCount:= 0;
  FCapacity:= 0;
  ReallocMem(FTokenList, 0 * SizeOf(PChar));
end;

destructor TTokenList.Destroy;
begin
  Clear;
  Inherited Destroy;
end;

procedure TTokenList.Expand;
begin
  SetCapacity(FCapacity + 1024);
end;

procedure TTokenList.SetCapacity(NewCapacity:Integer);
begin
  FCapacity:= NewCapacity;
  ReallocMem(FTokenList, FCapacity * SizeOf(PChar));
end;

function TTokenList.Add(Item: PChar):Integer;
begin
  Result := FCount;
  if Result = FCapacity then Expand;
  FTokenList[Result] := Item;
  Inc(FCount);
end;

function TTokenList.GetItems(Index: Integer):PChar;
begin
  Result := FTokenList[Index];
end;

procedure TTokenList.SetItems(Index: Integer;  Item: PChar);
begin
  FTokenList[Index]:= Item;
end;

procedure TTokenList.Pack;
begin
  ReallocMem(FTokenList, FCount * SizeOf(PChar));
end;

function TTokenList.GetLength(Index: Integer):Integer;
var
  RunPtr, StartPtr: PChar;
  TokenLen: Integer;
begin
  StartPtr:= FTokenList[Index];
  RunPtr:= StartPtr;
  if ScanStr(FDelimiters,RunPtr^)<> nil then
  begin
    inc(RunPtr);
  end else
  repeat
    if RunPtr^ = #0 then exit;
    Inc(RunPtr);
  until ScanStr(FDelimiters,RunPtr^)<> nil;
  Result:= RunPtr - StartPtr;
end;

function TTokenList.GetPosition(Index: Integer):Integer;
begin
  FTokenPos:= FTokenList[Index];
  Result:= FTokenPos - FOrigin;
end;

function TTokenList.GetToken(Index: Integer):String;
var
  RunPtr, StartPtr: PChar;
  TokenLen: Integer;
  aToken: array[0..1024] of Char;
begin
  StartPtr:= FTokenList[Index];
  TokenLen:= GetLength(Index);
  StrLCopy(aToken, StartPtr, TokenLen);
  aToken[TokenLen]:= #0;
  Result:= aToken;
end;

procedure TTokenList.SetDelimiters(value:String);
begin
  StrPCopy(FDelimiters, value);
end;

function TTokenList.GetDelimiters:String;
begin
  Result := FDelimiters;
end;

procedure TTokenList.Tokenize(StartCapacity: Integer);
begin
  SetCapacity(StartCapacity);
  Run:= FOrigin;
  while (Run^<>#0) do
  begin
    Start:= Run;
    Add(Start);
    if ScanStr(FDelimiters,Run^)<> nil then
    begin
      inc(Run);
    end else
    repeat
      if Run^ = #0 then exit;
      Inc(Run);
    until ScanStr(FDelimiters,Run^)<> nil;
  end;
end;

function TTokenList.Next:String;
var
  TokenLen: Integer;
  aToken: array[0..1024] of Char;
begin
  TokenLen:= 0;
  Start:= Run;
  if ScanStr(FDelimiters,Run^)<> nil then
  begin
    if Run^ <> #0 then begin Inc(Run); Inc(TokenLen) end;
  end else
  repeat
    if Run^ = #0 then exit;
    Inc(Run);
    Inc(TokenLen);
  until ScanStr(FDelimiters,Run^)<> nil;
  if Run^ = #0 then FFinished:= True;
  if TokenLen = 1 then Result:= Start^ else
  begin
    StrLCopy(aToken, Start, TokenLen);
    aToken[TokenLen]:= #0;
    Result:= aToken;
  end;
end;

{ Although StrScan is in Assembler this is faster, about 40%}
function TTokenList.ScanStr(ToScan: PChar; Sign: Char):PChar;
begin
  Result:= nil;
  if ToScan <> nil then
  while (ToScan^ <> #0) do
  begin
    if ToScan^ = Sign then
    begin
      Result:= ToScan;
      break;
    end;
    inc(ToScan);
  end;
end;

{ This is equal to that, what the Compiler does with it.
  I wasn't able to improve it in any way.
  Maybe somebody will try it.}
{function TTokenList.ScanStr(ToScan: PChar; Sign: Char):PChar;assembler;
asm
        PUSH    EBX
        //Result: = nil
        XOR     EBX,EBX
        //if ToScan <> nil then
        TEST    EDX,EDX
        JZ      @@4
        JMP     @@1
        //if ToScan^ = Sign then
@@2:    CMP     CL,AL
        JNZ     @@3
        //Result: = ToScan
        MOV     EBX,EDX
        //break;
        JMP     @@4
        //inc(ToScan)
@@3:    INC     EDX
        //while ToScan^ <> #0 do
@@1:    MOV     AL,[EDX]
        TEST    AL,AL
        JNZ     @@2
@@4:    MOV     EAX,EBX
        Pop     EBX
end;}

end.

{ Martin Waldenburg
  Landaeckerstrasse 27
  71642 Ludwigsburg
  Germany }
